import unittest

import gtirb


class TestProperties(unittest.TestCase):
    def test_data_blocks(self):
        b = gtirb.DataBlock()

        self.assertEquals(b.address, None)
        self.assertEquals(b.contents, b"")
        self.assertEquals(set(b.references), set())

        bi = gtirb.ByteInterval(address=1, contents=b"abcd1234")
        b.offset = 2
        b.size = 3
        b.byte_interval = bi

        self.assertEquals(b.address, 3)
        self.assertEquals(b.contents, b"cd1")
        self.assertEquals(set(b.references), set())

        s = gtirb.Section()
        bi.section = s
        m = gtirb.Module(name="M")
        sym1 = gtirb.Symbol("test", payload=b)
        sym2 = gtirb.Symbol("test", payload=123)
        sym3 = gtirb.Symbol("test", payload=b)
        m.symbols |= {sym1, sym2, sym3}
        s.module = m

        self.assertEquals(b.address, 3)
        self.assertEquals(b.contents, b"cd1")
        self.assertEquals(set(b.references), {sym1, sym3})

    def test_code_blocks(self):
        b = gtirb.CodeBlock()

        self.assertEquals(b.address, None)
        self.assertEquals(b.contents, b"")
        self.assertEquals(set(b.references), set())
        self.assertEquals(set(b.incoming_edges), set())
        self.assertEquals(set(b.outgoing_edges), set())

        bi = gtirb.ByteInterval(address=1, contents=b"abcd1234")
        b.offset = 2
        b.size = 3
        b.byte_interval = bi

        self.assertEquals(b.address, 3)
        self.assertEquals(b.contents, b"cd1")
        self.assertEquals(set(b.references), set())
        self.assertEquals(set(b.incoming_edges), set())
        self.assertEquals(set(b.outgoing_edges), set())

        s = gtirb.Section()
        bi.section = s
        m = gtirb.Module(name="M")
        sym1 = gtirb.Symbol("test", payload=b)
        sym2 = gtirb.Symbol("test", payload=123)
        sym3 = gtirb.Symbol("test", payload=b)
        m.symbols |= {sym1, sym2, sym3}
        s.module = m

        self.assertEquals(b.address, 3)
        self.assertEquals(b.contents, b"cd1")
        self.assertEquals(set(b.references), {sym1, sym3})
        self.assertEquals(set(b.incoming_edges), set())
        self.assertEquals(set(b.outgoing_edges), set())

        i = gtirb.IR()
        m.ir = i
        p1 = gtirb.ProxyBlock()
        p2 = gtirb.ProxyBlock()
        p3 = gtirb.ProxyBlock()
        p4 = gtirb.ProxyBlock()
        i.cfg.add(gtirb.Edge(b, p1))
        i.cfg.add(gtirb.Edge(p2, b))
        i.cfg.add(gtirb.Edge(p3, p4))
        i.cfg.add(gtirb.Edge(b, b))

        self.assertEquals(b.address, 3)
        self.assertEquals(b.contents, b"cd1")
        self.assertEquals(set(b.references), {sym1, sym3})
        self.assertEquals(
            set((s, t) for s, t, l in b.incoming_edges), {(p2, b), (b, b)}
        )
        self.assertEquals(
            set((s, t) for s, t, l in b.outgoing_edges), {(b, p1), (b, b)}
        )

    def test_proxy_blocks(self):
        b = gtirb.ProxyBlock()

        self.assertEquals(set(b.references), set())
        self.assertEquals(set(b.incoming_edges), set())
        self.assertEquals(set(b.outgoing_edges), set())

        m = gtirb.Module(name="M")
        sym1 = gtirb.Symbol("test", payload=b)
        sym2 = gtirb.Symbol("test", payload=123)
        sym3 = gtirb.Symbol("test", payload=b)
        m.symbols |= {sym1, sym2, sym3}
        b.module = m

        self.assertEquals(set(b.references), {sym1, sym3})
        self.assertEquals(set(b.incoming_edges), set())
        self.assertEquals(set(b.outgoing_edges), set())

        i = gtirb.IR()
        m.ir = i
        p1 = gtirb.ProxyBlock()
        p2 = gtirb.ProxyBlock()
        p3 = gtirb.ProxyBlock()
        p4 = gtirb.ProxyBlock()
        i.cfg.add(gtirb.Edge(b, p1))
        i.cfg.add(gtirb.Edge(p2, b))
        i.cfg.add(gtirb.Edge(p3, p4))
        i.cfg.add(gtirb.Edge(b, b))

        self.assertEquals(set(b.references), {sym1, sym3})
        self.assertEquals(
            set((s, t) for s, t, l in b.incoming_edges), {(p2, b), (b, b)}
        )
        self.assertEquals(
            set((s, t) for s, t, l in b.outgoing_edges), {(b, p1), (b, b)}
        )

    def test_sections(self):
        s = gtirb.Section()
        self.assertEquals(s.address, None)
        self.assertEquals(s.size, None)

        s.byte_intervals.clear()
        s.byte_intervals |= {gtirb.ByteInterval()}
        self.assertEquals(s.address, None)
        self.assertEquals(s.size, None)

        s.byte_intervals.clear()
        s.byte_intervals |= {gtirb.ByteInterval(size=3)}
        self.assertEquals(s.address, None)
        self.assertEquals(s.size, None)

        s.byte_intervals.clear()
        s.byte_intervals |= {gtirb.ByteInterval(address=2, size=4)}
        self.assertEquals(s.address, 2)
        self.assertEquals(s.size, 4)

        s.byte_intervals.clear()
        s.byte_intervals |= {
            gtirb.ByteInterval(address=2, size=4),
            gtirb.ByteInterval(size=3),
        }
        self.assertEquals(s.address, None)
        self.assertEquals(s.size, None)

        s.byte_intervals.clear()
        s.byte_intervals |= {
            gtirb.ByteInterval(address=2, size=4),
            gtirb.ByteInterval(address=100, size=3),
        }
        self.assertEquals(s.address, 2)
        self.assertEquals(s.size, 101)

    def test_modules(self):
        s1 = gtirb.Section(
            name="s1", byte_intervals=[gtirb.ByteInterval(address=4, size=4)]
        )
        s2 = gtirb.Section(
            name="s2", byte_intervals=[gtirb.ByteInterval(address=8, size=8)]
        )
        s3 = gtirb.Section(
            name="s3", byte_intervals=[gtirb.ByteInterval(address=100, size=1)]
        )
        s4 = gtirb.Section(
            name="s4", byte_intervals=[gtirb.ByteInterval(size=1000)]
        )
        m = gtirb.Module(name="M", sections=[s1, s2, s3, s4])

        self.assertEquals(set(m.sections_on(3)), set())
        self.assertEquals(set(m.sections_on(4)), {s1})
        self.assertEquals(set(m.sections_on(7)), {s1})
        self.assertEquals(set(m.sections_on(8)), {s2})
        self.assertEquals(set(m.sections_on(15)), {s2})
        self.assertEquals(set(m.sections_on(16)), set())
        self.assertEquals(set(m.sections_on(99)), set())
        self.assertEquals(set(m.sections_on(100)), {s3})
        self.assertEquals(set(m.sections_on(101)), set())

        self.assertEquals(set(m.sections_on(range(0, 100))), {s1, s2})
        self.assertEquals(set(m.sections_on(range(0, 101))), {s1, s2, s3})
        self.assertEquals(set(m.sections_on(range(0, 102))), {s1, s2, s3})
        self.assertEquals(set(m.sections_on(range(7, 7 + 4))), {s1, s2})
        self.assertEquals(set(m.sections_on(range(8, 8 + 4))), {s2})
        self.assertEquals(set(m.sections_on(range(17, 17 + 80))), set())

        self.assertEquals(set(m.sections_at(3)), set())
        self.assertEquals(set(m.sections_at(4)), {s1})
        self.assertEquals(set(m.sections_at(5)), set())
        self.assertEquals(set(m.sections_at(7)), set())
        self.assertEquals(set(m.sections_at(8)), {s2})
        self.assertEquals(set(m.sections_at(9)), set())
        self.assertEquals(set(m.sections_at(99)), set())
        self.assertEquals(set(m.sections_at(100)), {s3})
        self.assertEquals(set(m.sections_at(101)), set())

        self.assertEquals(set(m.sections_at(range(0, 100))), {s1, s2})
        self.assertEquals(set(m.sections_at(range(0, 101))), {s1, s2, s3})
        self.assertEquals(set(m.sections_at(range(5, 10))), {s2})
        self.assertEquals(set(m.sections_at(range(95, 105))), {s3})

    def test_byte_intervals(self):
        s = gtirb.Symbol(name="test")
        se1 = gtirb.SymAddrConst(0, s)
        se3 = gtirb.SymAddrAddr(0, 1, s, s)
        bi = gtirb.ByteInterval(
            address=10, size=5, symbolic_expressions={0: se1, 4: se3}
        )

        self.assertEqual(set(bi.symbolic_expressions_at(9)), set())
        self.assertEqual(set(bi.symbolic_expressions_at(10)), {(bi, 0, se1)})
        self.assertEqual(set(bi.symbolic_expressions_at(11)), set())
        self.assertEqual(set(bi.symbolic_expressions_at(13)), set())
        self.assertEqual(set(bi.symbolic_expressions_at(14)), {(bi, 4, se3)})
        self.assertEqual(set(bi.symbolic_expressions_at(15)), set())

        self.assertEqual(set(bi.symbolic_expressions_at(range(0, 9))), set())
        self.assertEqual(set(bi.symbolic_expressions_at(range(11, 14))), set())
        self.assertEqual(set(bi.symbolic_expressions_at(range(20, 90))), set())
        self.assertEqual(
            set(bi.symbolic_expressions_at(range(0, 90))),
            {(bi, 0, se1), (bi, 4, se3)},
        )
        self.assertEqual(
            set(bi.symbolic_expressions_at(range(10, 15))),
            {(bi, 0, se1), (bi, 4, se3)},
        )
        self.assertEqual(
            set(bi.symbolic_expressions_at(range(11, 18))), {(bi, 4, se3)}
        )

    def test_sym_exprs(self):
        node = gtirb.SymAddrConst(
            offset=123,
            symbol=gtirb.Symbol(name="symbol", payload=gtirb.ProxyBlock()),
        )
        self.assertEqual({x.name for x in node.symbols}, {"symbol"})

        node = gtirb.SymAddrAddr(
            offset=123,
            scale=2,
            symbol1=gtirb.Symbol(name="symbol1", payload=gtirb.ProxyBlock()),
            symbol2=gtirb.Symbol(name="symbol2", payload=gtirb.ProxyBlock()),
        )
        self.assertEqual(
            {x.name for x in node.symbols}, {"symbol1", "symbol2"}
        )



import unittest

import gtirb


class ModuleTests(unittest.TestCase):
    def test_symbols_named(self):
        m = gtirb.Module(name="test")
        s1 = gtirb.Symbol(name="hello", module=m)
        s2 = gtirb.Symbol(name="world", module=m)

        found = set(m.symbols_named("hello"))
        self.assertEqual(found, {s1})

        found = set(m.symbols_named("world"))
        self.assertEqual(found, {s2})

        # Change the name to verify we update the index
        s1.name = "world"
        found = set(m.symbols_named("hello"))
        self.assertEqual(found, set())

        found = set(m.symbols_named("world"))
        self.assertEqual(found, {s1, s2})

        # Discard the symbol to verify we update the index
        m.symbols.discard(s1)
        found = set(m.symbols_named("world"))
        self.assertEqual(found, {s2})

        # Now add it back to verify we update the index
        m.symbols.add(s1)
        found = set(m.symbols_named("world"))
        self.assertEqual(found, {s1, s2})



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gtirb/proto/SymbolicExpression.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='gtirb/proto/SymbolicExpression.proto',
  package='gtirb.proto',
  syntax='proto3',
  serialized_options=_b('\n\032com.grammatech.gtirb.proto'),
  serialized_pb=_b('\n$gtirb/proto/SymbolicExpression.proto\x12\x0bgtirb.proto\"4\n\rSymStackConst\x12\x0e\n\x06offset\x18\x01 \x01(\x05\x12\x13\n\x0bsymbol_uuid\x18\x02 \x01(\x0c\"3\n\x0cSymAddrConst\x12\x0e\n\x06offset\x18\x01 \x01(\x03\x12\x13\n\x0bsymbol_uuid\x18\x02 \x01(\x0c\"X\n\x0bSymAddrAddr\x12\r\n\x05scale\x18\x01 \x01(\x03\x12\x0e\n\x06offset\x18\x02 \x01(\x03\x12\x14\n\x0csymbol1_uuid\x18\x03 \x01(\x0c\x12\x14\n\x0csymbol2_uuid\x18\x04 \x01(\x0c\"\xc4\x01\n\x12SymbolicExpression\x12/\n\naddr_const\x18\x02 \x01(\x0b\x32\x19.gtirb.proto.SymAddrConstH\x00\x12-\n\taddr_addr\x18\x03 \x01(\x0b\x32\x18.gtirb.proto.SymAddrAddrH\x00\x12\x32\n\x0f\x61ttribute_flags\x18\x04 \x03(\x0e\x32\x19.gtirb.proto.SymAttributeB\x07\n\x05valueJ\x04\x08\x01\x10\x02R\x0bstack_const*\xec\x05\n\x0cSymAttribute\x12\x07\n\x03GOT\x10\x00\x12\t\n\x05GOTPC\x10\x01\x12\n\n\x06GOTOFF\x10\x02\x12\n\n\x06GOTREL\x10\x03\x12\x07\n\x03PLT\x10\x04\x12\n\n\x06PLTOFF\x10\x05\x12\t\n\x05PCREL\x10\x06\x12\n\n\x06SECREL\x10\x07\x12\x07\n\x03TLS\x10\x08\x12\t\n\x05TLSGD\x10\t\x12\t\n\x05TLSLD\x10\n\x12\n\n\x06TLSLDM\x10\x0b\x12\x0b\n\x07TLSCALL\x10\x0c\x12\x0b\n\x07TLSDESC\x10\r\x12\t\n\x05TPREL\x10\x0e\x12\t\n\x05TPOFF\x10\x0f\x12\n\n\x06\x44TPREL\x10\x10\x12\n\n\x06\x44TPOFF\x10\x11\x12\n\n\x06NTPOFF\x10\x12\x12\n\n\x06\x44TPMOD\x10\x13\x12\x08\n\x04PAGE\x10\x14\x12\x0b\n\x07PAGEOFF\x10\x15\x12\x08\n\x04\x43\x41LL\x10\x16\x12\x06\n\x02LO\x10\x17\x12\x06\n\x02HI\x10\x18\x12\n\n\x06HIGHER\x10\x19\x12\x0b\n\x07HIGHEST\x10\x1a\x12\x0e\n\tGOTNTPOFF\x10\xe8\x07\x12\x0e\n\tINDNTPOFF\x10\xe9\x07\x12\x07\n\x02G0\x10\xd1\x0f\x12\x07\n\x02G1\x10\xd2\x0f\x12\x07\n\x02G2\x10\xd3\x0f\x12\x07\n\x02G3\x10\xd4\x0f\x12\x0c\n\x07UPPER16\x10\xd5\x0f\x12\x0c\n\x07LOWER16\x10\xd6\x0f\x12\t\n\x04LO12\x10\xd7\x0f\x12\t\n\x04LO15\x10\xd8\x0f\x12\t\n\x04LO14\x10\xd9\x0f\x12\t\n\x04HI12\x10\xda\x0f\x12\t\n\x04HI21\x10\xdb\x0f\x12\x06\n\x01S\x10\xdc\x0f\x12\x07\n\x02PG\x10\xdd\x0f\x12\x07\n\x02NC\x10\xde\x0f\x12\x08\n\x03\x41\x42S\x10\xdf\x0f\x12\t\n\x04PREL\x10\xe0\x0f\x12\x0b\n\x06PREL31\x10\xe1\x0f\x12\x0c\n\x07TARGET1\x10\xe2\x0f\x12\x0c\n\x07TARGET2\x10\xe3\x0f\x12\n\n\x05SBREL\x10\xe4\x0f\x12\x0b\n\x06TLSLDO\x10\xe5\x0f\x12\t\n\x04HI16\x10\xb8\x17\x12\t\n\x04LO16\x10\xb9\x17\x12\n\n\x05GPREL\x10\xba\x17\x12\t\n\x04\x44ISP\x10\xbb\x17\x12\t\n\x04OFST\x10\xbc\x17\x12\x06\n\x01H\x10\xa0\x1f\x12\x06\n\x01L\x10\xa1\x1f\x12\x07\n\x02HA\x10\xa2\x1f\x12\t\n\x04HIGH\x10\xa3\x1f\x12\n\n\x05HIGHA\x10\xa4\x1f\x12\x0c\n\x07HIGHERA\x10\xa5\x1f\x12\r\n\x08HIGHESTA\x10\xa6\x1f\x12\x0c\n\x07TOCBASE\x10\xa7\x1f\x12\x08\n\x03TOC\x10\xa8\x1f\x12\n\n\x05NOTOC\x10\xa9\x1f\x42\x1c\n\x1a\x63om.grammatech.gtirb.protob\x06proto3')
)

_SYMATTRIBUTE = _descriptor.EnumDescriptor(
  name='SymAttribute',
  full_name='gtirb.proto.SymAttribute',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GOT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GOTPC', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GOTOFF', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GOTREL', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PLT', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PLTOFF', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PCREL', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SECREL', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TLS', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TLSGD', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TLSLD', index=10, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TLSLDM', index=11, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TLSCALL', index=12, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TLSDESC', index=13, number=13,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TPREL', index=14, number=14,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TPOFF', index=15, number=15,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DTPREL', index=16, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DTPOFF', index=17, number=17,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NTPOFF', index=18, number=18,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DTPMOD', index=19, number=19,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAGE', index=20, number=20,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAGEOFF', index=21, number=21,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CALL', index=22, number=22,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LO', index=23, number=23,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HI', index=24, number=24,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HIGHER', index=25, number=25,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HIGHEST', index=26, number=26,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GOTNTPOFF', index=27, number=1000,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INDNTPOFF', index=28, number=1001,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='G0', index=29, number=2001,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='G1', index=30, number=2002,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='G2', index=31, number=2003,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='G3', index=32, number=2004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UPPER16', index=33, number=2005,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOWER16', index=34, number=2006,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LO12', index=35, number=2007,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LO15', index=36, number=2008,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LO14', index=37, number=2009,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HI12', index=38, number=2010,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HI21', index=39, number=2011,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='S', index=40, number=2012,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PG', index=41, number=2013,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NC', index=42, number=2014,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ABS', index=43, number=2015,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PREL', index=44, number=2016,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PREL31', index=45, number=2017,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TARGET1', index=46, number=2018,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TARGET2', index=47, number=2019,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SBREL', index=48, number=2020,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TLSLDO', index=49, number=2021,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HI16', index=50, number=3000,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LO16', index=51, number=3001,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GPREL', index=52, number=3002,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISP', index=53, number=3003,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OFST', index=54, number=3004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='H', index=55, number=4000,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='L', index=56, number=4001,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HA', index=57, number=4002,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HIGH', index=58, number=4003,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HIGHA', index=59, number=4004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HIGHERA', index=60, number=4005,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HIGHESTA', index=61, number=4006,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOCBASE', index=62, number=4007,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOC', index=63, number=4008,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOTOC', index=64, number=4009,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=450,
  serialized_end=1198,
)
_sym_db.RegisterEnumDescriptor(_SYMATTRIBUTE)

SymAttribute = enum_type_wrapper.EnumTypeWrapper(_SYMATTRIBUTE)
GOT = 0
GOTPC = 1
GOTOFF = 2
GOTREL = 3
PLT = 4
PLTOFF = 5
PCREL = 6
SECREL = 7
TLS = 8
TLSGD = 9
TLSLD = 10
TLSLDM = 11
TLSCALL = 12
TLSDESC = 13
TPREL = 14
TPOFF = 15
DTPREL = 16
DTPOFF = 17
NTPOFF = 18
DTPMOD = 19
PAGE = 20
PAGEOFF = 21
CALL = 22
LO = 23
HI = 24
HIGHER = 25
HIGHEST = 26
GOTNTPOFF = 1000
INDNTPOFF = 1001
G0 = 2001
G1 = 2002
G2 = 2003
G3 = 2004
UPPER16 = 2005
LOWER16 = 2006
LO12 = 2007
LO15 = 2008
LO14 = 2009
HI12 = 2010
HI21 = 2011
S = 2012
PG = 2013
NC = 2014
ABS = 2015
PREL = 2016
PREL31 = 2017
TARGET1 = 2018
TARGET2 = 2019
SBREL = 2020
TLSLDO = 2021
HI16 = 3000
LO16 = 3001
GPREL = 3002
DISP = 3003
OFST = 3004
H = 4000
L = 4001
HA = 4002
HIGH = 4003
HIGHA = 4004
HIGHERA = 4005
HIGHESTA = 4006
TOCBASE = 4007
TOC = 4008
NOTOC = 4009



_SYMSTACKCONST = _descriptor.Descriptor(
  name='SymStackConst',
  full_name='gtirb.proto.SymStackConst',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='offset', full_name='gtirb.proto.SymStackConst.offset', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol_uuid', full_name='gtirb.proto.SymStackConst.symbol_uuid', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=105,
)


_SYMADDRCONST = _descriptor.Descriptor(
  name='SymAddrConst',
  full_name='gtirb.proto.SymAddrConst',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='offset', full_name='gtirb.proto.SymAddrConst.offset', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol_uuid', full_name='gtirb.proto.SymAddrConst.symbol_uuid', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=158,
)


_SYMADDRADDR = _descriptor.Descriptor(
  name='SymAddrAddr',
  full_name='gtirb.proto.SymAddrAddr',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='scale', full_name='gtirb.proto.SymAddrAddr.scale', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset', full_name='gtirb.proto.SymAddrAddr.offset', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol1_uuid', full_name='gtirb.proto.SymAddrAddr.symbol1_uuid', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol2_uuid', full_name='gtirb.proto.SymAddrAddr.symbol2_uuid', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=160,
  serialized_end=248,
)


_SYMBOLICEXPRESSION = _descriptor.Descriptor(
  name='SymbolicExpression',
  full_name='gtirb.proto.SymbolicExpression',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='addr_const', full_name='gtirb.proto.SymbolicExpression.addr_const', index=0,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addr_addr', full_name='gtirb.proto.SymbolicExpression.addr_addr', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attribute_flags', full_name='gtirb.proto.SymbolicExpression.attribute_flags', index=2,
      number=4, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='gtirb.proto.SymbolicExpression.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=251,
  serialized_end=447,
)

_SYMBOLICEXPRESSION.fields_by_name['addr_const'].message_type = _SYMADDRCONST
_SYMBOLICEXPRESSION.fields_by_name['addr_addr'].message_type = _SYMADDRADDR
_SYMBOLICEXPRESSION.fields_by_name['attribute_flags'].enum_type = _SYMATTRIBUTE
_SYMBOLICEXPRESSION.oneofs_by_name['value'].fields.append(
  _SYMBOLICEXPRESSION.fields_by_name['addr_const'])
_SYMBOLICEXPRESSION.fields_by_name['addr_const'].containing_oneof = _SYMBOLICEXPRESSION.oneofs_by_name['value']
_SYMBOLICEXPRESSION.oneofs_by_name['value'].fields.append(
  _SYMBOLICEXPRESSION.fields_by_name['addr_addr'])
_SYMBOLICEXPRESSION.fields_by_name['addr_addr'].containing_oneof = _SYMBOLICEXPRESSION.oneofs_by_name['value']
DESCRIPTOR.message_types_by_name['SymStackConst'] = _SYMSTACKCONST
DESCRIPTOR.message_types_by_name['SymAddrConst'] = _SYMADDRCONST
DESCRIPTOR.message_types_by_name['SymAddrAddr'] = _SYMADDRADDR
DESCRIPTOR.message_types_by_name['SymbolicExpression'] = _SYMBOLICEXPRESSION
DESCRIPTOR.enum_types_by_name['SymAttribute'] = _SYMATTRIBUTE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SymStackConst = _reflection.GeneratedProtocolMessageType('SymStackConst', (_message.Message,), dict(
  DESCRIPTOR = _SYMSTACKCONST,
  __module__ = 'gtirb.proto.SymbolicExpression_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.SymStackConst)
  ))
_sym_db.RegisterMessage(SymStackConst)

SymAddrConst = _reflection.GeneratedProtocolMessageType('SymAddrConst', (_message.Message,), dict(
  DESCRIPTOR = _SYMADDRCONST,
  __module__ = 'gtirb.proto.SymbolicExpression_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.SymAddrConst)
  ))
_sym_db.RegisterMessage(SymAddrConst)

SymAddrAddr = _reflection.GeneratedProtocolMessageType('SymAddrAddr', (_message.Message,), dict(
  DESCRIPTOR = _SYMADDRADDR,
  __module__ = 'gtirb.proto.SymbolicExpression_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.SymAddrAddr)
  ))
_sym_db.RegisterMessage(SymAddrAddr)

SymbolicExpression = _reflection.GeneratedProtocolMessageType('SymbolicExpression', (_message.Message,), dict(
  DESCRIPTOR = _SYMBOLICEXPRESSION,
  __module__ = 'gtirb.proto.SymbolicExpression_pb2'
  # @@protoc_insertion_point(class_scope:gtirb.proto.SymbolicExpression)
  ))
_sym_db.RegisterMessage(SymbolicExpression)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import unittest

import gtirb
from helpers import SearchScope, parameterize_one


class ByteIntervalsOnTests(unittest.TestCase):
    @parameterize_one(
        "scope", (SearchScope.ir, SearchScope.module, SearchScope.section)
    )
    def test_byte_intervals_on(self, scope):
        ir = gtirb.IR()
        m = gtirb.Module(name="test", ir=ir)
        s = gtirb.Section(module=m)
        search_in = scope.select(ir, m, s, None)

        bi1 = gtirb.ByteInterval(address=0x1000, size=4, section=s)
        bi2 = gtirb.ByteInterval(address=0x1004, size=4, section=s)

        found = set(search_in.byte_intervals_on(0x1000))
        self.assertEqual(found, {bi1})

        found = set(search_in.byte_intervals_on(0x1001))
        self.assertEqual(found, {bi1})

        found = set(search_in.byte_intervals_on(range(0x1000, 0x1008)))
        self.assertEqual(found, {bi1, bi2})

        found = set(search_in.byte_intervals_on(range(0x1000, 0x1008, 16)))
        self.assertEqual(found, {bi1, bi2})

        # Change the address to verify we update the index
        bi2.address = 0x2000

        found = set(search_in.byte_intervals_on(0x1005))
        self.assertEqual(found, set())

        found = set(search_in.byte_intervals_on(0x2001))
        self.assertEqual(found, {bi2})

        # Discard the interval to verify we update the index
        bi2.section = None

        found = set(search_in.byte_intervals_on(0x2001))
        self.assertEqual(found, set())

        # Now add it back to verify we update the index
        s.byte_intervals.add(bi2)
        found = set(search_in.byte_intervals_on(0x2001))
        self.assertEqual(found, {bi2})

    @parameterize_one(
        "scope", [SearchScope.ir, SearchScope.module, SearchScope.section]
    )
    def test_byte_intervals_overlapping(self, scope):
        ir = gtirb.IR()
        m = gtirb.Module(name="test", ir=ir)
        s = gtirb.Section(module=m)
        search_in = scope.select(ir, m, s, None)

        bi1 = gtirb.ByteInterval(address=0x1000, size=8, section=s)
        bi2 = gtirb.ByteInterval(address=0x1004, size=4, section=s)

        found = set(search_in.byte_intervals_on(0x1005))
        self.assertEqual(found, {bi1, bi2})



import typing
from enum import Enum
from uuid import UUID

from .node import Node
from .proto import SymbolicExpression_pb2
from .symbol import Symbol
from .util import DeserializationError

AttributesCtorType = typing.Iterable[
    typing.Union[
        "SymbolicExpression_pb2.SymAttribute.ValueType",
        "SymbolicExpression.Attribute",
    ]
]


class SymbolicExpression:
    """Base class of symbolic expression types."""

    class Attribute(Enum):
        """Attributes representing a known property of a symbolic expression.
        See https://grammatech.github.io/gtirb/md__symbolic_expression.html
        """

        GOT = SymbolicExpression_pb2.SymAttribute.Value("GOT")
        GOTPC = SymbolicExpression_pb2.SymAttribute.Value("GOTPC")
        GOTOFF = SymbolicExpression_pb2.SymAttribute.Value("GOTOFF")
        GOTREL = SymbolicExpression_pb2.SymAttribute.Value("GOTREL")
        PLT = SymbolicExpression_pb2.SymAttribute.Value("PLT")
        PLTOFF = SymbolicExpression_pb2.SymAttribute.Value("PLTOFF")
        PCREL = SymbolicExpression_pb2.SymAttribute.Value("PCREL")
        SECREL = SymbolicExpression_pb2.SymAttribute.Value("SECREL")
        TLS = SymbolicExpression_pb2.SymAttribute.Value("TLS")
        TLSGD = SymbolicExpression_pb2.SymAttribute.Value("TLSGD")
        TLSLD = SymbolicExpression_pb2.SymAttribute.Value("TLSLD")
        TLSLDM = SymbolicExpression_pb2.SymAttribute.Value("TLSLDM")
        TLSCALL = SymbolicExpression_pb2.SymAttribute.Value("TLSCALL")
        TLSDESC = SymbolicExpression_pb2.SymAttribute.Value("TLSDESC")
        TPREL = SymbolicExpression_pb2.SymAttribute.Value("TPREL")
        TPOFF = SymbolicExpression_pb2.SymAttribute.Value("TPOFF")
        DTPREL = SymbolicExpression_pb2.SymAttribute.Value("DTPREL")
        DTPOFF = SymbolicExpression_pb2.SymAttribute.Value("DTPOFF")
        DTPMOD = SymbolicExpression_pb2.SymAttribute.Value("DTPMOD")
        NTPOFF = SymbolicExpression_pb2.SymAttribute.Value("NTPOFF")
        PAGE = SymbolicExpression_pb2.SymAttribute.Value("PAGE")
        PAGEOFF = SymbolicExpression_pb2.SymAttribute.Value("PAGEOFF")
        CALL = SymbolicExpression_pb2.SymAttribute.Value("CALL")
        LO = SymbolicExpression_pb2.SymAttribute.Value("LO")
        HI = SymbolicExpression_pb2.SymAttribute.Value("HI")
        HIGHER = SymbolicExpression_pb2.SymAttribute.Value("HIGHER")
        HIGHEST = SymbolicExpression_pb2.SymAttribute.Value("HIGHEST")
        GOTNTPOFF = SymbolicExpression_pb2.SymAttribute.Value("GOTNTPOFF")
        INDNTPOFF = SymbolicExpression_pb2.SymAttribute.Value("INDNTPOFF")
        G0 = SymbolicExpression_pb2.SymAttribute.Value("G0")
        G1 = SymbolicExpression_pb2.SymAttribute.Value("G1")
        G2 = SymbolicExpression_pb2.SymAttribute.Value("G2")
        G3 = SymbolicExpression_pb2.SymAttribute.Value("G3")
        UPPER16 = SymbolicExpression_pb2.SymAttribute.Value("UPPER16")
        LOWER16 = SymbolicExpression_pb2.SymAttribute.Value("LOWER16")
        LO12 = SymbolicExpression_pb2.SymAttribute.Value("LO12")
        LO15 = SymbolicExpression_pb2.SymAttribute.Value("LO15")
        LO14 = SymbolicExpression_pb2.SymAttribute.Value("LO14")
        HI12 = SymbolicExpression_pb2.SymAttribute.Value("HI12")
        HI21 = SymbolicExpression_pb2.SymAttribute.Value("HI21")
        S = SymbolicExpression_pb2.SymAttribute.Value("S")
        PG = SymbolicExpression_pb2.SymAttribute.Value("PG")
        NC = SymbolicExpression_pb2.SymAttribute.Value("NC")
        ABS = SymbolicExpression_pb2.SymAttribute.Value("ABS")
        PREL = SymbolicExpression_pb2.SymAttribute.Value("PREL")
        PREL31 = SymbolicExpression_pb2.SymAttribute.Value("PREL31")
        TARGET1 = SymbolicExpression_pb2.SymAttribute.Value("TARGET1")
        TARGET2 = SymbolicExpression_pb2.SymAttribute.Value("TARGET2")
        SBREL = SymbolicExpression_pb2.SymAttribute.Value("SBREL")
        TLSLDO = SymbolicExpression_pb2.SymAttribute.Value("TLSLDO")
        HI16 = SymbolicExpression_pb2.SymAttribute.Value("HI16")
        LO16 = SymbolicExpression_pb2.SymAttribute.Value("LO16")
        GPREL = SymbolicExpression_pb2.SymAttribute.Value("GPREL")
        DISP = SymbolicExpression_pb2.SymAttribute.Value("DISP")
        OFST = SymbolicExpression_pb2.SymAttribute.Value("OFST")
        H = SymbolicExpression_pb2.SymAttribute.Value("H")
        L = SymbolicExpression_pb2.SymAttribute.Value("L")
        HA = SymbolicExpression_pb2.SymAttribute.Value("HA")
        HIGH = SymbolicExpression_pb2.SymAttribute.Value("HIGH")
        HIGHA = SymbolicExpression_pb2.SymAttribute.Value("HIGHA")
        HIGHERA = SymbolicExpression_pb2.SymAttribute.Value("HIGHERA")
        HIGHESTA = SymbolicExpression_pb2.SymAttribute.Value("HIGHESTA")
        TOCBASE = SymbolicExpression_pb2.SymAttribute.Value("TOCBASE")
        TOC = SymbolicExpression_pb2.SymAttribute.Value("TOC")
        NOTOC = SymbolicExpression_pb2.SymAttribute.Value("NOTOC")

        def __repr__(self) -> str:
            return "SymbolicExpression.Attribute.%s" % self.name

        def __int__(self) -> int:
            return self.value

    def __init__(
        self,
        attributes: AttributesCtorType = set(),
    ):
        self.attributes = set(attributes)

    @property
    def symbols(self) -> typing.Iterable[Symbol]:
        """Get all the symbols involved with this symbolic expression,
        regardless of role.
        """

        raise NotImplementedError  # pragma: no cover

    def deep_eq(self, other: object) -> bool:
        raise NotImplementedError  # pragma: no cover

    def _attributes_repr(self) -> str:
        if not self.attributes:
            return "set()"
        else:
            return "{%s}" % ",".join(repr(a) for a in self.attributes)


class SymAddrAddr(SymbolicExpression):
    """Represents a symbolic expression of the form
    "(Sym1 - Sym2) / Scale + Offset".

    :ivar ~.scale: Constant scale factor.
    :ivar ~.offset: Constant offset.
    :ivar ~.symbol1: Symbol representing the base address.
    :ivar ~.symbol2: Symbol to subtract from ``symbol1``.
    """

    def __init__(
        self,
        scale: int,
        offset: int,
        symbol1: Symbol,
        symbol2: Symbol,
        attributes: AttributesCtorType = set(),
    ):
        """
        :param scale: Constant scale factor.
        :param offset: Constant offset.
        :param symbol1: Symbol representing the base address.
        :param symbol2: Symbol to subtract from ``symbol1``.
        :param attributes: :class:`SymobolicExpression.Attribute`\\s of this
            expression.
        """
        super().__init__(attributes)
        self.scale = scale
        self.offset = offset
        self.symbol1 = symbol1
        self.symbol2 = symbol2

    @classmethod
    def _from_protobuf(
        cls,
        proto_symaddraddr: SymbolicExpression_pb2.SymAddrAddr,
        get_by_uuid: typing.Callable[[UUID], typing.Optional[Node]],
    ) -> "SymAddrAddr":
        symbol1_uuid = UUID(bytes=proto_symaddraddr.symbol1_uuid)
        symbol1 = get_by_uuid(symbol1_uuid)
        if not isinstance(symbol1, Symbol):
            raise DeserializationError(
                "SymAddrAddr: UUID %s is not a Symbol" % symbol1_uuid
            )
        symbol2_uuid = UUID(bytes=proto_symaddraddr.symbol2_uuid)
        symbol2 = get_by_uuid(symbol2_uuid)
        if not isinstance(symbol2, Symbol):
            raise DeserializationError(
                "SymAddrAddr: UUID %s is not a Symbol" % symbol2_uuid
            )
        return cls(
            proto_symaddraddr.scale, proto_symaddraddr.offset, symbol1, symbol2
        )

    def _to_protobuf(self) -> SymbolicExpression_pb2.SymAddrAddr:
        proto_symaddraddr = SymbolicExpression_pb2.SymAddrAddr()
        proto_symaddraddr.scale = self.scale
        proto_symaddraddr.offset = self.offset
        proto_symaddraddr.symbol1_uuid = self.symbol1.uuid.bytes
        proto_symaddraddr.symbol2_uuid = self.symbol2.uuid.bytes
        return proto_symaddraddr

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SymAddrAddr):
            return False
        return (
            self.scale == other.scale
            and self.offset == other.offset
            and self.symbol1.uuid == other.symbol1.uuid
            and self.symbol2.uuid == other.symbol2.uuid
            and self.attributes == other.attributes
        )

    def __hash__(self) -> int:
        return hash(
            (self.offset, self.scale, self.symbol1.uuid, self.symbol2.uuid)
        )

    def __repr__(self) -> str:
        return (
            "SymAddrAddr("
            "scale={scale!r}, "
            "offset={offset!r}, "
            "symbol1={symbol1!r}, "
            "symbol2={symbol2!r}, "
            "attributes={attributes_repr!s}, "
            ")"
        ).format(
            scale=self.scale,
            offset=self.offset,
            symbol1=self.symbol1,
            symbol2=self.symbol2,
            attributes_repr=self._attributes_repr(),
        )

    def deep_eq(self, other: object) -> bool:
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, SymAddrAddr):
            return False
        return (
            self.scale == other.scale
            and self.offset == other.offset
            and self.symbol1.deep_eq(other.symbol1)
            and self.symbol2.deep_eq(other.symbol2)
            and self.attributes == other.attributes
        )

    @property
    def symbols(self) -> typing.Iterable[Symbol]:
        yield self.symbol1
        yield self.symbol2


class SymAddrConst(SymbolicExpression):
    """Represents a symbolic expression of the form "Sym + Offset".

    :ivar ~.offset: Constant offset.
    :ivar ~.symbol: Symbol representing an address.
    """

    def __init__(
        self,
        offset: int,
        symbol: Symbol,
        attributes: AttributesCtorType = set(),
    ):
        """
        :param offset: Constant offset.
        :param symbol: Symbol representing an address.
        :param attributes: :class:`SymbolicExpression.Attribute`\\s of this
            expression.
        """
        super().__init__(attributes)
        self.offset = offset
        self.symbol = symbol

    @classmethod
    def _from_protobuf(
        cls,
        proto_symaddrconst: SymbolicExpression_pb2.SymAddrConst,
        get_by_uuid: typing.Callable[[UUID], typing.Optional[Node]],
    ) -> "SymAddrConst":
        symbol_uuid = UUID(bytes=proto_symaddrconst.symbol_uuid)
        symbol = get_by_uuid(symbol_uuid)
        if not isinstance(symbol, Symbol):
            raise DeserializationError(
                "SymAddrConst: UUID %s is not a Symbol" % symbol_uuid
            )
        return cls(proto_symaddrconst.offset, symbol)

    def _to_protobuf(self) -> SymbolicExpression_pb2.SymAddrConst:
        proto_symaddrconst = SymbolicExpression_pb2.SymAddrConst()
        proto_symaddrconst.offset = self.offset
        if self.symbol is not None:
            proto_symaddrconst.symbol_uuid = self.symbol.uuid.bytes
        return proto_symaddrconst

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SymAddrConst):
            return False
        return (
            self.offset == other.offset
            and self.symbol.uuid == other.symbol.uuid
            and self.attributes == other.attributes
        )

    def __hash__(self) -> int:
        return hash((self.offset, self.symbol.uuid))

    def __repr__(self) -> str:
        return (
            "SymAddrConst("
            "offset={offset!r}, "
            "symbol={symbol!r}, "
            "attributes={attributes_repr!s}, "
            ")"
        ).format(
            offset=self.offset,
            symbol=self.symbol,
            attributes_repr=self._attributes_repr(),
        )

    def deep_eq(self, other: object) -> bool:
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, SymAddrConst):
            return False
        return (
            self.offset == other.offset
            and self.symbol.deep_eq(other.symbol)
            and self.attributes == other.attributes
        )

    @property
    def symbols(self) -> typing.Iterable[Symbol]:
        yield self.symbol



"""isort:skip_file"""

import unittest
import gtirb
import uuid

# imports needed because repr's use of unqualified names
from uuid import UUID  # noqa: F401,F403,F401
from gtirb import *  # noqa: F401,F403,F401


class ReprTest(unittest.TestCase):
    def test_auxdata(self):
        node = gtirb.AuxData(
            type_name="mapping<string,set<UUID>>",
            data={
                "a": set([uuid.uuid4(), uuid.uuid4()]),
                "b": set([uuid.uuid4()]),
            },
        )
        string = repr(node)
        new_node = eval(string)
        # auxdata has no deep_eq
        # (because how can we ensure "data" has a deep_eq?)
        self.assertEqual(node.type_name, new_node.type_name)
        self.assertEqual(node.data, new_node.data)

    def test_block(self):
        node = gtirb.CodeBlock(
            offset=123, size=456, decode_mode=gtirb.CodeBlock.DecodeMode.Thumb
        )
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_proxy_block(self):
        node = gtirb.ProxyBlock()
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_data_object(self):
        node = gtirb.DataBlock(offset=123, size=456)
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_ir(self):
        # TODO: expand this
        node = gtirb.IR()
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_cfg(self):
        node = gtirb.CFG()
        node.add(
            gtirb.Edge(
                gtirb.CodeBlock(offset=1, size=2),
                gtirb.CodeBlock(offset=3, size=4),
                gtirb.Edge.Label(
                    type=gtirb.Edge.Type.Fallthrough,
                    conditional=True,
                    direct=False,
                ),
            )
        )
        node.add(
            gtirb.Edge(
                gtirb.CodeBlock(offset=5, size=6),
                gtirb.CodeBlock(offset=7, size=8),
                gtirb.Edge.Label(
                    type=gtirb.Edge.Type.Branch,
                    conditional=True,
                    direct=False,
                ),
            )
        )
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_module(self):
        # TODO: expand this
        node = gtirb.Module(name="M")
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_offset(self):
        node = gtirb.Offset(element_id=uuid.uuid4(), displacement=123)
        string = repr(node)
        new_node = eval(string)
        self.assertEqual(node.element_id, new_node.element_id)
        self.assertEqual(node.displacement, new_node.displacement)

    def test_section(self):
        node = gtirb.Section(
            name=".text",
            flags=(gtirb.Section.Flag.Readable, gtirb.Section.Flag.Writable),
        )
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_symbol(self):
        node = gtirb.Symbol(name="symbol1", payload=gtirb.ProxyBlock())
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

        node = gtirb.Symbol(name="symbol2", payload=0x123)
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

        node = gtirb.Symbol(name="symbol3", payload=None)
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_sym_expr(self):
        node = gtirb.SymAddrConst(
            offset=123,
            symbol=gtirb.Symbol(name="symbol", payload=gtirb.ProxyBlock()),
            attributes=set(),
        )
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

        node = gtirb.SymAddrAddr(
            offset=123,
            scale=2,
            symbol1=gtirb.Symbol(name="symbol1", payload=gtirb.ProxyBlock()),
            symbol2=gtirb.Symbol(name="symbol2", payload=gtirb.ProxyBlock()),
            attributes={
                gtirb.SymbolicExpression.Attribute.ABS,
                gtirb.SymbolicExpression.Attribute.G0,
            },
        )
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))

    def test_byte_interval(self):
        node = gtirb.ByteInterval(
            address=0x123,
            initialized_size=456,
            size=789,
            contents=b"abc",
            blocks=(gtirb.DataBlock(size=0),),
            symbolic_expressions={
                1: gtirb.SymAddrConst(offset=1, symbol=gtirb.Symbol("test"))
            },
        )
        string = repr(node)
        new_node = eval(string)
        self.assertTrue(node.deep_eq(new_node))


if __name__ == "__main__":
    unittest.main()



import typing
from enum import Enum
from uuid import UUID

from .node import Node, _NodeMessage
from .proto import CodeBlock_pb2, DataBlock_pb2, ProxyBlock_pb2
from .util import _IndexedAttribute

if typing.TYPE_CHECKING:  # pragma: no cover
    # Ignore flake8 "imported but unused" errors.
    from .byteinterval import ByteInterval  # noqa: F401
    from .cfg import Edge  # noqa: F401
    from .ir import IR  # noqa: F401
    from .module import Module  # noqa: F401
    from .section import Section  # noqa: F401
    from .symbol import Symbol  # noqa: F401


class Block(Node):
    """The base class for blocks. Symbols may have references to any subclass
    of Block.
    """

    @property
    def references(self) -> typing.Iterator["Symbol"]:
        """Get all the symbols that refer to this block."""

        raise NotImplementedError  # pragma: no cover

    def _add_to_uuid_cache(self, cache: typing.Dict[UUID, Node]) -> None:
        """Update the UUID cache when this node is added."""

        cache[self.uuid] = self

    def _remove_from_uuid_cache(self, cache: typing.Dict[UUID, Node]) -> None:
        """Update the UUID cache when this node is removed."""

        del cache[self.uuid]


class ByteBlock(Block):
    """The base class for blocks that belong to a :class:`ByteInterval` and
    store their bytes there.

    :ivar ~.size: The size of the block in bytes.
    :ivar ~.offset: The offset from the beginning of the byte interval to which
        this block belongs. Multiple blocks in the same interval may have the
        same offset.
    """

    size = _IndexedAttribute[int]()(lambda self: self.byte_interval)
    offset = _IndexedAttribute[int]()(lambda self: self.byte_interval)

    def __init__(
        self,
        *,
        size: int = 0,
        offset: int = 0,
        uuid: typing.Optional[UUID] = None,
        byte_interval: typing.Optional["ByteInterval"] = None,
    ):
        """
        :param size: The size of the data object in bytes.
        :param offset: The offset from the beginning of the byte interval to
            which this block belongs.
        :param uuid: The UUID of this ``ByteBlock``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        :param byte_interval: The :class:`ByteInterval` this block belongs to.
        """

        super().__init__(uuid=uuid)
        self._byte_interval: typing.Optional["ByteInterval"] = None
        self.size = size
        self.offset = offset
        # Use the property setter to ensure correct invariants.
        self.byte_interval = byte_interval

    @property
    def byte_interval(self) -> typing.Optional["ByteInterval"]:
        """The :class:`ByteInterval` this block belongs to."""

        return self._byte_interval

    @byte_interval.setter
    def byte_interval(self, value: typing.Optional["ByteInterval"]) -> None:
        if self._byte_interval is not None:
            self._byte_interval.blocks.discard(self)
        if value is not None:
            value.blocks.add(self)

    def deep_eq(self, other: object) -> bool:
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, ByteBlock):
            return False
        return (
            self.offset == other.offset
            and self.uuid == other.uuid
            and self.size == other.size
        )

    @property
    def contents(self) -> bytes:
        """Get the bytes in this block."""

        if self.byte_interval is None:
            return b""
        return self.byte_interval.contents[
            self.offset : self.offset + self.size
        ]

    @property
    def address(self) -> typing.Optional[int]:
        """Get the address of this block, or None if not present."""

        if self.byte_interval is None or self.byte_interval.address is None:
            return None
        return self.byte_interval.address + self.offset

    @property
    def references(self) -> typing.Iterator["Symbol"]:
        if not self.module:
            return

        symbol_set = self.module._symbol_referent_index.get(self)
        if symbol_set:
            yield from symbol_set

    @property
    def section(self) -> typing.Optional["Section"]:
        """Get the section this node ultimately belongs to."""
        if self.byte_interval is None:
            return None
        return self.byte_interval.section

    @property
    def module(self) -> typing.Optional["Module"]:
        """Get the module this node ultimately belongs to."""
        if self.section is None:
            return None
        return self.section.module

    @property
    def ir(self) -> typing.Optional["IR"]:
        """Get the IR this node ultimately belongs to."""
        if self.module is None:
            return None
        return self.module.ir

    def contains_offset(self, offset: int) -> bool:
        """Indicate if the provided offset is within this block."""
        return self.offset <= offset < (self.offset + self.size)

    def contains_address(self, address: int) -> bool:
        """Indicate if the provided address is within this block.
        Returns False if the block has no address.
        """
        byte_interval = self.byte_interval
        if byte_interval is not None:
            base = byte_interval.address
            if base is not None:
                return self.contains_offset(address - base)
        return False


class CfgNode(Block):
    """The base class for blocks that may appear as vertices in the CFG."""

    @property
    def incoming_edges(self) -> typing.Iterable["Edge"]:
        """Get the edges that point to this CFG node."""

        raise NotImplementedError  # pragma: no cover

    @property
    def outgoing_edges(self) -> typing.Iterable["Edge"]:
        """Get the edges that start at this CFG node."""

        raise NotImplementedError  # pragma: no cover


class DataBlock(ByteBlock):
    """Represents a data object, possibly symbolic."""

    def __init__(
        self,
        *,
        size: int = 0,
        offset: int = 0,
        uuid: typing.Optional[UUID] = None,
        byte_interval: typing.Optional["ByteInterval"] = None,
    ):
        """
        :param size: The size of the data object in bytes.
        :param offset: The offset from the beginning of the byte interval to
            which this block belongs.
        :param uuid: The UUID of this ``DataBlock``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        :param byte_interval: The :class:`ByteInterval` this block belongs to.
        """

        super().__init__(
            size=size, offset=offset, uuid=uuid, byte_interval=byte_interval
        )

    @classmethod
    def _decode_protobuf(
        cls,
        proto_dataobject: _NodeMessage,
        uuid: UUID,
        ir: typing.Optional["IR"],
    ) -> "DataBlock":
        assert ir
        assert isinstance(proto_dataobject, DataBlock_pb2.DataBlock)
        b = cls(size=proto_dataobject.size, uuid=uuid)
        b._add_to_uuid_cache(ir._local_uuid_cache)
        return b

    def _to_protobuf(self) -> DataBlock_pb2.DataBlock:
        proto_dataobject = DataBlock_pb2.DataBlock()
        proto_dataobject.uuid = self.uuid.bytes
        proto_dataobject.size = self.size
        return proto_dataobject

    def __repr__(self) -> str:
        return (
            "DataBlock("
            "uuid={uuid!r}, "
            "size={size}, "
            "offset={offset}, "
            ")".format(uuid=self.uuid, size=self.size, offset=self.offset)
        )


class CodeBlock(ByteBlock, CfgNode):
    """A basic block in the binary.

    Does not directly store data bytes, which are kept in a
    :class:`ByteInterval`.

    :ivar ~.decode_mode: The decode mode of the block,
        used in some ISAs to differentiate between sub-ISAs
        (e.g. differentiating blocks written in ARM and Thumb).
    """

    class DecodeMode(Enum):
        """Variations on decoding a particular ISA"""

        Default = CodeBlock_pb2.DecodeMode.Value("All_Default")
        """Default decode mode for all ISAs"""

        Thumb = CodeBlock_pb2.DecodeMode.Value("ARM_Thumb")
        """Thumb decode mode for ARM32"""

    def __init__(
        self,
        *,
        decode_mode: DecodeMode = DecodeMode.Default,
        size: int = 0,
        offset: int = 0,
        uuid: typing.Optional[UUID] = None,
        byte_interval: typing.Optional["ByteInterval"] = None,
    ):
        """
        :param size: The length of the block in bytes.
        :param decode_mode: The decode mode of the block,
            used in some ISAs to differentiate between sub-ISAs
            (e.g. differentiating blocks written in ARM and Thumb).
            Defaults to DecodeMode.Default.
        :param offset: The offset from the beginning of the byte interval to
            which this block belongs.
        :param uuid: The UUID of this ``CodeBlock``,
            or None if a new UUID needs generated via :func:`uuid.uuid4`.
            Defaults to None.
        :param byte_interval: The :class:`ByteInterval` this block belongs to.
        """

        super().__init__(
            size=size, offset=offset, uuid=uuid, byte_interval=byte_interval
        )
        self.decode_mode = decode_mode  # type: CodeBlock.DecodeMode

    @classmethod
    def _decode_protobuf(
        cls,
        proto_block: _NodeMessage,
        uuid: UUID,
        ir: typing.Optional["IR"],
    ) -> "CodeBlock":
        assert ir
        assert isinstance(proto_block, CodeBlock_pb2.CodeBlock)
        b = cls(
            decode_mode=cls.DecodeMode(proto_block.decode_mode),
            size=proto_block.size,
            uuid=uuid,
        )
        b._add_to_uuid_cache(ir._local_uuid_cache)
        return b

    def _to_protobuf(self) -> CodeBlock_pb2.CodeBlock:
        proto_block = CodeBlock_pb2.CodeBlock()
        proto_block.uuid = self.uuid.bytes
        proto_block.size = self.size
        proto_block.decode_mode = self.decode_mode.value
        return proto_block

    def deep_eq(self, other: object) -> bool:
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, CodeBlock):
            return False
        return super().deep_eq(other) and self.decode_mode == other.decode_mode

    def __repr__(self) -> str:
        return (
            "CodeBlock("
            "uuid={uuid!r}, "
            "size={size}, "
            "offset={offset}, "
            "decode_mode=CodeBlock.{decode_mode!s}, "
            ")".format(
                uuid=self.uuid,
                size=self.size,
                offset=self.offset,
                decode_mode=self.decode_mode,
            )
        )

    @property
    def incoming_edges(self) -> typing.Iterator["Edge"]:
        if self.ir is None:
            return iter(())
        return self.ir.cfg.in_edges(self)

    @property
    def outgoing_edges(self) -> typing.Iterator["Edge"]:
        if self.ir is None:
            return iter(())
        return self.ir.cfg.out_edges(self)


class ProxyBlock(CfgNode):
    """A placeholder that serves as the endpoint (source or target) of a
    :class:`gtirb.Edge`.

    ProxyBlock objects allow the construction of CFG edges to or from
    another node. For example, a call to a function in another module
    may be represented by a :class:`gtirb.Edge` that originates at the
    calling :class:`gtirb.CodeBlock` and targets a ProxyBlock. Another
    example would be a :class:`gtirb.Edge` that represents an indirect
    jump whose target is not known.

    A ProxyBlock does not represent any instructions and so has neither
    an address nor a size.
    """

    def __init__(
        self,
        *,
        uuid: typing.Optional[UUID] = None,
        module: typing.Optional["Module"] = None,
    ):
        super().__init__(uuid=uuid)
        self._module: typing.Optional["Module"] = None
        # Use the property setter to ensure correct invariants.
        self.module = module

    @classmethod
    def _decode_protobuf(
        cls,
        proto_proxy: _NodeMessage,
        uuid: UUID,
        ir: typing.Optional["IR"],
    ) -> "ProxyBlock":
        assert ir
        assert isinstance(proto_proxy, ProxyBlock_pb2.ProxyBlock)
        b = cls(uuid=uuid)
        b._add_to_uuid_cache(ir._local_uuid_cache)
        return b

    def _to_protobuf(self) -> ProxyBlock_pb2.ProxyBlock:
        proto_proxyblock = ProxyBlock_pb2.ProxyBlock()
        proto_proxyblock.uuid = self.uuid.bytes
        return proto_proxyblock

    def deep_eq(self, other: object) -> bool:
        # Do not move __eq__. See docstring for Node.deep_eq for more info.
        if not isinstance(other, ProxyBlock):
            return False
        return self.uuid == other.uuid

    def __repr__(self) -> str:
        return "ProxyBlock(" "uuid={uuid!r}, " ")".format(**self.__dict__)

    @property
    def module(self) -> typing.Optional["Module"]:
        return self._module

    @module.setter
    def module(self, value: typing.Optional["Module"]) -> None:
        if self._module is not None:
            self._module.proxies.discard(self)
        if value is not None:
            value.proxies.add(self)

    @property
    def references(self) -> typing.Iterator["Symbol"]:
        if self.module is None:
            return iter(())
        return (s for s in self.module.symbols if s.referent == self)

    @property
    def incoming_edges(self) -> typing.Iterator["Edge"]:
        if self.ir is None:
            return iter(())
        return self.ir.cfg.in_edges(self)

    @property
    def outgoing_edges(self) -> typing.Iterator["Edge"]:
        if self.ir is None:
            return iter(())
        return self.ir.cfg.out_edges(self)

    @property
    def ir(self) -> typing.Optional["IR"]:
        """Get the IR this node ultimately belongs to."""
        if self.module is None:
            return None
        return self.module.ir



import io
import os
import tempfile
import unittest

import gtirb

IR_FILE = tempfile.mktemp(suffix=".gtirb")


class IRTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        ir = gtirb.IR()
        m = gtirb.Module(
            binary_path="binary_path",
            file_format=gtirb.Module.FileFormat.RAW,
            isa=gtirb.Module.ISA.ValidButUnsupported,
            name="name",
            preferred_addr=1,
            rebase_delta=2,
            ir=ir,
        )
        s = gtirb.Section(
            name="name",
            flags=(
                gtirb.Section.Flag.Executable,
                gtirb.Section.Flag.Readable,
                gtirb.Section.Flag.Loaded,
                gtirb.Section.Flag.Initialized,
            ),
            module=m,
        )
        bi = gtirb.ByteInterval(
            address=0, size=10, contents=b"abcd", section=s
        )
        cb = gtirb.CodeBlock(
            size=4,
            offset=0,
            decode_mode=gtirb.CodeBlock.DecodeMode.Thumb,
            byte_interval=bi,
        )
        _ = gtirb.DataBlock(size=6, offset=4, byte_interval=bi)
        sym = gtirb.Symbol(name="name", payload=cb, module=m)
        sac = gtirb.SymAddrConst(
            0, sym, {gtirb.SymbolicExpression.Attribute.G1}
        )
        bi.symbolic_expressions[2] = sac
        p = gtirb.ProxyBlock(module=m)
        ir.cfg.add(
            gtirb.Edge(
                cb,
                p,
                gtirb.Edge.Label(
                    type=gtirb.Edge.Type.Branch, conditional=False, direct=True
                ),
            )
        )
        ir.cfg.add(gtirb.Edge(p, p))
        m.aux_data["key"] = gtirb.AuxData(gtirb.Offset(s, 777), "Offset")
        ir.aux_data["key"] = gtirb.AuxData("value", "string")

        self.ir = ir

    def setUp(self):
        self.ir.save_protobuf(IR_FILE)

    def tearDown(self):
        os.remove(IR_FILE)

    def test_ir_protobuf_load(self):
        new_ir = gtirb.IR.load_protobuf(IR_FILE)
        self.assertTrue(self.ir.deep_eq(new_ir))
        self.assertNotEqual(
            self.ir.modules[0].aux_data["key"].data,
            new_ir.modules[0].aux_data["key"].data,
        )


class NotGTIRBTest(unittest.TestCase):
    def test(self):
        file_content = io.BytesIO(b"JUNK")
        with self.assertRaises(Exception) as context:
            gtirb.IR.load_protobuf_file(file_content)

        self.assertEqual(
            "File missing GTIRB magic - not a GTIRB file?",
            str(context.exception),
        )


class BadVersionTest(unittest.TestCase):
    def test(self):
        file_content = io.BytesIO(b"GTIRB\x00\x00\xFF")
        with self.assertRaises(Exception) as context:
            gtirb.IR.load_protobuf_file(file_content)

        self.assertTrue(
            "Attempt to decode IR of version" in str(context.exception)
        )


class BadProtobufTest(unittest.TestCase):
    def test(self):
        bytes = b"GTIRB\x00\x00"
        bytes += gtirb.version.PROTOBUF_VERSION.to_bytes(1, byteorder="little")
        bytes += b"JUNK"
        file_content = io.BytesIO(bytes)
        with self.assertRaises(Exception) as context:
            gtirb.IR.load_protobuf_file(file_content)

        self.assertTrue("Error parsing message" in str(context.exception))


if __name__ == "__main__":
    unittest.main()



import unittest

import gtirb
from helpers import SearchScope, create_interval_etc, parameterize_one


class SymbolicExpressionsAtTests(unittest.TestCase):
    @parameterize_one("scope", list(SearchScope))
    def test_symbolic_expressions_at(self, scope):
        ir, m, s, bi = create_interval_etc(address=0x1000, size=5)
        search_in = scope.select(ir, m, s, bi)

        sym = gtirb.Symbol(name="hello")
        expr = gtirb.SymAddrConst(0, sym)
        bi.symbolic_expressions[1] = expr
        bi.symbolic_expressions[2] = expr
        bi.symbolic_expressions[3] = expr
        bi.symbolic_expressions[4] = expr

        found = set(search_in.symbolic_expressions_at(0x1001))
        self.assertEqual(found, {(bi, 1, expr)})

        found = set(search_in.symbolic_expressions_at(range(0x1000, 0x1004)))
        self.assertEqual(found, {(bi, 1, expr), (bi, 2, expr), (bi, 3, expr)})

        found = set(
            search_in.symbolic_expressions_at(range(0x1000, 0x1004, 2))
        )
        self.assertEqual(found, {(bi, 2, expr)})

        # Now just verify that the index updates correctly when deleting
        del bi.symbolic_expressions[1]

        found = set(search_in.symbolic_expressions_at(0x1001))
        self.assertEqual(found, set())



# Get the version from the package's version, so the version can be defined in
# setup.py alone.
API_VERSION = (
    "1."
    "11."
    "0"
    ""
)  # type: str
"""The semantic version of this API."""

PROTOBUF_VERSION = 4  # type: int
"""The version of Protobuf this API can read and write from.
Attempts to load old Protobuf versions will raise a ``ValueError``.
"""



import unittest

import gtirb.util


class RecordingList(gtirb.util.ListWrapper):
    def __init__(self, *args):
        self.record = []
        super().__init__(*args)

    def _add(self, value):
        self.record.append(("add", value))

    def _remove(self, value):
        self.record.append(("remove", value))


class ListWrapperTests(unittest.TestCase):
    def test_empty_init(self):
        lst = RecordingList()
        self.assertFalse(lst.record)

    def test_values_init(self):
        lst = RecordingList(["A", "B", "C"])
        self.assertEqual(
            lst.record, [("add", "A"), ("add", "B"), ("add", "C")]
        )

    def test_getitem_int(self):
        lst = RecordingList(["A", "B", "C"])
        self.assertEqual(lst[0], "A")
        self.assertEqual(lst[1], "B")
        self.assertEqual(lst[2], "C")

    def test_getitem_slice(self):
        lst = RecordingList(["A", "B", "C"])
        self.assertEqual(lst[1:2], ["B"])
        self.assertEqual(lst[:2], ["A", "B"])
        self.assertEqual(lst[1:], ["B", "C"])

    def test_setitem_int(self):
        lst = RecordingList(["A", "B", "C"])
        lst[1] = "D"
        self.assertEqual(list(lst), ["A", "D", "C"])
        self.assertEqual(
            lst.record,
            [
                ("add", "A"),
                ("add", "B"),
                ("add", "C"),
                ("remove", "B"),
                ("add", "D"),
            ],
        )

    def test_setitem_slice(self):
        lst = RecordingList(["A", "B", "C"])
        lst[1:2] = ["D", "E"]
        self.assertEqual(list(lst), ["A", "D", "E", "C"])
        self.assertEqual(
            lst.record,
            [
                ("add", "A"),
                ("add", "B"),
                ("add", "C"),
                ("remove", "B"),
                ("add", "D"),
                ("add", "E"),
            ],
        )

    def test_delitem_int(self):
        lst = RecordingList(["A", "B"])
        del lst[0]
        self.assertEqual(list(lst), ["B"])
        self.assertEqual(
            lst.record, [("add", "A"), ("add", "B"), ("remove", "A")]
        )

    def test_delitem_slice(self):
        lst = RecordingList(["A", "B", "C"])
        del lst[1:]
        self.assertEqual(list(lst), ["A"])
        self.assertEqual(
            lst.record,
            [
                ("add", "A"),
                ("add", "B"),
                ("add", "C"),
                ("remove", "B"),
                ("remove", "C"),
            ],
        )

    def test_len(self):
        lst = RecordingList(["A", "B", "C"])
        self.assertEqual(len(lst), 3)

    def test_insert(self):
        lst = RecordingList(["A", "B"])
        lst.insert(1, "C")
        self.assertEqual(list(lst), ["A", "C", "B"])
        self.assertEqual(
            lst.record, [("add", "A"), ("add", "B"), ("add", "C")]
        )

    def test_append(self):
        lst = RecordingList(["A", "B"])
        lst.append("C")
        self.assertEqual(list(lst), ["A", "B", "C"])
        self.assertEqual(
            lst.record, [("add", "A"), ("add", "B"), ("add", "C")]
        )

    def test_remove(self):
        lst = RecordingList(["A", "B"])
        lst.remove("A")
        self.assertEqual(list(lst), ["B"])
        self.assertEqual(
            lst.record, [("add", "A"), ("add", "B"), ("remove", "A")]
        )

    def test_extend(self):
        lst = RecordingList()
        lst.extend(["A", "B"])
        self.assertEqual(list(lst), ["A", "B"])
        self.assertEqual(lst.record, [("add", "A"), ("add", "B")])

    def test_str(self):
        lst = RecordingList(["A", "B"])
        self.assertEqual(str(lst), str(["A", "B"]))
